/*
 * This file is part of the Configuration File Library (CFL).
 * Copyright (C) 2002-2003 Nicolas Darnis <ndarnis@free.fr>
 *
 * The CFL library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The CFL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the CFL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: cfl.c
 * $Revision: 1.9 $
 * $Date: 2003/05/28 15:53:51 $
 * $Author: ndarnis $
 * $Id: cfl.c,v 1.9 2003/05/28 15:53:51 ndarnis Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <config.h>

#include <gdsl.h>

#include "cfl.h"
#include "parser.h"
#include "section.h"

struct _cfl_handle
{
  gdsl_list_t sections;
};

struct error_message 
{
  int   id;
  char *msgid;
  char *msg;
};

static struct error_message ErrorMessages[] = 
{
  {CFL_ERR_MALLOC            , "CFL_ERR_MALLOC"            , "Memory allocation error"},
  {CFL_ERR_SECTION_EXISTS    , "CFL_ERR_SECTION_EXISTS"    , "The section already exists"},
  {CFL_ERR_SECTION_NOT_EXISTS, "CFL_ERR_SECTION_NOT_EXISTS", "The section does not exists"},
  {CFL_ERR_VALUE_EXISTS      , "CFL_ERR_VALUE_EXISTS"      , "The value already exists"},
  {CFL_ERR_VALUE_NOT_EXISTS  , "CFL_ERR_VALUE_NOT_EXIST"   , "The value does not exists"}
};

#define ERR_MSG_NB (sizeof (ErrorMessages) / sizeof (struct error_message) + 1)

/******************************************************************************/
/* Public functions                                                           */
/******************************************************************************/

extern cflh_t
cfl_alloc (void)
{
  cflh_t handle;

  handle = (cflh_t) malloc (sizeof (struct _cfl_handle));

  if (handle == NULL)
    {
      return NULL;
    }

  handle->sections = gdsl_list_create ("SECTIONS", section_alloc, section_free);

  if (handle->sections == NULL)
    {
      free (handle);
      return NULL;
    }

  /* We always create a MAIN nameless section */
  if (gdsl_list_insert_tail (handle->sections, NULL) == NULL)
    {
      gdsl_list_destroy (handle->sections);
      free (handle);
      return NULL;
    }
  
  return handle;
}

extern void
cfl_free (cflh_t handle)
{
  gdsl_list_destroy (handle->sections);
  free (handle);
}

extern cflh_t
cfl_read (const char *input_file_name)
{ 
  cflh_t handle;
  FILE * file;

  file = fopen (input_file_name, "r");

  if (file == NULL)
    {
      return NULL;
    }

  handle = cfl_alloc ();

  if (handle == NULL)
    {
      fclose (file);
      return NULL;
    }

  if (cfl_init_from_file (handle, file, CFL_FILE_TYPE_DEFAULT) < 0)
    {
      cfl_free (handle); 
      fclose (file);
      return NULL; 
    }

  fclose (file);

  return handle;
}

extern int
cfl_init_from_file (cflh_t handle, const FILE *input_file, const cfl_file_type_t file_type)
{
  /* there is more than the unique nameless section */
  if (gdsl_list_get_card (handle->sections) > 1) 
    {
      gdsl_list_flush (handle->sections); /* flush ALL sections */

      /* we re-create the nameless section */
      if (gdsl_list_insert_tail (handle->sections, NULL) == NULL)
	{
	  return CFL_ERR_MALLOC;
	}
    }
  
  return yy_parse_file (handle, input_file, file_type);
}

extern void
cfl_write (const cflh_t handle, FILE *output_file)
{
  gdsl_list_write (handle->sections, section_write, output_file, NULL);
}

extern void
cfl_write_xml (const cflh_t handle, FILE *output_file)
{
  fprintf (output_file, "<CFL REF=\"%p\" SECTIONS=\"%d\">\n", handle, cfl_get_sections_count (handle));
  gdsl_list_write (handle->sections, section_write_xml, output_file, NULL);
  fprintf (output_file, "</CFL>\n");
}

extern int
cfl_section_insert (cflh_t handle, const char *section_name)
{
  if (gdsl_list_search_by_value (handle->sections, section_compare, 
				 (void *) section_name) != NULL)
    {
      return CFL_ERR_SECTION_EXISTS;
    }

  if (gdsl_list_insert_tail (handle->sections, (void *) section_name) == NULL)
    {
      return CFL_ERR_MALLOC;
    }

  return 0;
}

extern int
cfl_section_remove (cflh_t handle, const char *section_name)
{
  cfl_section_t s;

  if (section_name == NULL)
    {
      return CFL_ERR_SECTION_NOT_EXISTS;
    }

  s = (cfl_section_t) gdsl_list_remove_by_value (handle->sections, 
						 section_compare, 
						 (void *) section_name);
  if(s == NULL)
    {
      return CFL_ERR_SECTION_NOT_EXISTS;
    }

  section_free (s);

  return 0;
}

extern int
cfl_get_sections_count (cflh_t handle)
{
  return gdsl_list_get_card (handle->sections);
}

extern char *
cfl_section_get_name_by_position (cflh_t handle, unsigned long pos)
{
  cfl_section_t s;

  s = gdsl_list_search_by_position (handle->sections, pos);

  if (s == NULL)
    {
      return NULL;
    }

  return section_get_name (s);
}

extern long
cfl_section_get_values_count (const cflh_t handle, const char *section_name) 
{
  cfl_section_t s;

  s = (cfl_section_t) gdsl_list_search_by_value (handle->sections, 
						 section_compare, 
						 (void *) section_name);

  if (s == NULL)
    {
      return CFL_ERR_SECTION_NOT_EXISTS;
    }

  return (long) section_get_values_number (s);
}

extern int
cfl_value_insert (cflh_t handle, const char *section_name, 
		  const char *value_name, const cfl_value_type_t type, void *val)
{
  int           rc;
  cfl_section_t s = NULL;

  s = (cfl_section_t) gdsl_list_search_by_value (handle->sections, 
						 section_compare, 
						 (void *) section_name);

  if (s == NULL)
    {
      return CFL_ERR_SECTION_NOT_EXISTS;
    }

  rc = section_add_value (s, value_name, type, val);

  if (rc < 0)
    {
      return rc;
    }
  
  return 0;
}

extern int
cfl_value_remove (cflh_t handle, const char *section_name, 
		  const char *value_name)
{
  int rc;
  gdsl_element_t e;

  e = gdsl_list_search_by_value (handle->sections, 
				 section_compare, 
				 (void *) section_name);

  if (e == NULL)
    {
      return CFL_ERR_SECTION_NOT_EXISTS;
    }

  rc = section_remove_value ((cfl_section_t) e, value_name);

  if (rc < 0) 
    {
      return rc;
    }

  return 0;
}

extern int
cfl_value_set_by_name (cflh_t handle, const char *section_name, 
		       const char *value_name, const cfl_value_type_t type, void *val)
{
  int rc;
  cfl_section_t s;

  s = (cfl_section_t) gdsl_list_search_by_value (handle->sections, 
						 section_compare, 
						 (void *) section_name);

  if (s == NULL)
    {
      return CFL_ERR_SECTION_NOT_EXISTS;
    }

  rc = section_set_value (s, value_name, type, val);

  if (rc < 0)
    {
      return rc;
    }
  
  return 0;
}

extern void *
cfl_value_get_by_name (cflh_t handle, const char *section_name, 
		       const char *value_name, cfl_value_type_t *type)
{
  cfl_section_t s;

  s = (cfl_section_t) gdsl_list_search_by_value (handle->sections, 
						 section_compare,
						 (void *) section_name);

  if (s == NULL)
    {
      return NULL;
    }

  return section_get_value (s, value_name, type);
}

extern void *
cfl_value_get_by_position (const cflh_t handle, const char *section_name, 
			   unsigned long pos, char **value_name, 
			   cfl_value_type_t *type)
{
  cfl_section_t s;

  s = (cfl_section_t) gdsl_list_search_by_value (handle->sections, 
						 section_compare, 
						 (void *) section_name);

  if (s == NULL)
    {
      return NULL;
    }

  return section_get_value_by_pos (s, pos, value_name, type);
}

extern const char *
cfl_errcode_to_string (int errcode)
{
  register int i;

  for (i = 0; i < ERR_MSG_NB; i++)
    {
      if (errcode == ErrorMessages[i].id)
	{
	  return ErrorMessages[i].msgid;
	}
    }

  return NULL;
}

extern const char *
cfl_errcode_to_message (int errcode)
{
  register int i;

  for (i = 0; i < ERR_MSG_NB; i++)
    {
      if (errcode == ErrorMessages[i].id)
	{
	  return ErrorMessages[i].msg;
	}
    }

  return NULL;
}

