/*
 * This file is part of the Configuration File Library (CFL).
 * Copyright (C) 2002-2003 Nicolas Darnis <ndarnis@free.fr>
 *
 * The CFL library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The CFL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the CFL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: parser.y
 * $Revision: 1.9 $
 * $Date: 2003/05/28 15:53:51 $
 * $Author: ndarnis $
 * $Id: parser.y,v 1.9 2003/05/28 15:53:51 ndarnis Exp $
 */

%{
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include <gdsl.h>

#include "cfl.h"
#include "parser.h"
#include "scanner.h"

static char * section_name = NULL;
static cflh_t handle = NULL;
static int    rc = 0;

static void
yy_error (const char *msg);
%}

%union 
{
  bool *    boolean;
  char *    string;
  char *    identifier;
  long int *integer;
  double *  real;
}

%token <boolean>    BOOLEAN
%token <string>     STRING
%token <integer>    INTEGER
%token <real>       REAL
%token <identifier> IDENTIFIER

%token AFFECT SECTION_OPEN SECTION_CLOSE
%type <identifier> ident

%start config

%%
config:        section_body
               | section_body sections_list
               ;

sections_list:   one_section
               | sections_list one_section
               ;

one_section:   SECTION_OPEN IDENTIFIER SECTION_CLOSE 
               { 
		 if ($2 == NULL) YYABORT;
		 if (section_name != NULL) {free (section_name); section_name = NULL;}
                 section_name = $2;
		 if (cfl_section_insert (handle, section_name) < 0) YYABORT;
               }
               section_body
               ;

section_body:  /* empty */
               | section_line section_body
               ;

section_line:  ident AFFECT STRING 
               { 
		 if ($1 == NULL || $3 == NULL) YYABORT;
                 rc = cfl_value_insert (handle, section_name, $1, CFL_TYPE_STRING, $3); 
		 free ($1);
		 free ($3);
                 if (rc < 0) YYABORT;
               }
               | ident AFFECT INTEGER 
               {
		 if ($1 == NULL || $3 == NULL) YYABORT;
                 rc = cfl_value_insert (handle, section_name, $1, CFL_TYPE_INTEGER, $3);
		 free ($1);
		 free ($3);
                 if (rc < 0) YYABORT;
               }
               | ident AFFECT REAL 
               { 
		 if ($1 == NULL || $3 == NULL) YYABORT;
                 rc = cfl_value_insert (handle, section_name, $1, CFL_TYPE_FLOAT, $3);
		 free ($1);
		 free ($3);
                 if (rc < 0) YYABORT;
               }
               | ident AFFECT BOOLEAN
               {
                 if ($1 == NULL || $3 == NULL) YYABORT;
                 rc = cfl_value_insert (handle, section_name, $1, CFL_TYPE_BOOL, (void *) $3);
                 free ($1);
		 free ($3);
                 if (rc < 0) YYABORT;
               }
               ;

ident:         IDENTIFIER
               ;
%%

/******************************************************************************/
/* Public functions                                                           */
/******************************************************************************/

extern int
yy_parse_file (cflh_t h, const FILE *input_file, const cfl_file_type_t file_type)
{
  int         rc;
  void *      buf;
  struct stat input_file_stat;

  if (fstat (fileno ((FILE *) input_file), &input_file_stat) < 0)
    {
      return CFL_ERR_FILE_READ;
    }

  buf = yy_start_lex (input_file, input_file_stat.st_size, file_type);
  if (buf == NULL)
    {
      return CFL_ERR_MALLOC;
    }

  handle = h;
  rc = yy_parse ();

  if (section_name != NULL)
    {
      free (section_name);
      section_name = NULL;
    }

  yy_stop_lex (buf);

  return (rc == 0) ? 0 : CFL_ERR_FILE_PARSE;
}

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static void
yy_error (const char *s)
{
#ifdef DEBUG
  fprintf (stderr, "ERROR: %s\n", s);
  fflush (stderr);
#endif
}
