%{
/*
 * This file is part of the Configuration File Library (CFL).
 * Copyright (C) 2002-2003 Nicolas Darnis <ndarnis@free.fr>
 *
 * The CFL library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The CFL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the CFL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: scanner.l
 * $Revision: 1.9 $
 * $Date: 2003/05/28 15:53:51 $
 * $Author: ndarnis $
 * $Id: scanner.l,v 1.9 2003/05/28 15:53:51 ndarnis Exp $
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <gdsl.h>

#include "cfl.h"
#include "scanner.h"
#include "parser.h"
#include "parser.tab.h"

static int make_string (void);
static int make_identifier (void);
static int make_integer (void);
static int make_real (void);
static int make_boolean (bool val);
%}

%option noyywrap 

space                   [ \t]
sign                    [-+]
letter                  [a-zA-Z_.-]
digit                   [0-9]

string                  [^"\n]*
qstring                 \"{string}["\n]
id                      {letter}({letter}|{digit})*
identifier              {id}({space}+{id})*
ident                   {identifier}("["{digit}+"]"{identifier}?)?
integer                 {sign}?{digit}+
real                    {integer}?\.{digit}+

%x affect

%%
"#".*                   ;
";".*                   ;
{space}+                ;
{ident}                 return make_identifier ();

"["                     return SECTION_OPEN;
"]"                     return SECTION_CLOSE;
"="                     BEGIN (affect); return AFFECT;
.|\n                    ;

<affect>"#".*           ;
<affect>";".*           ;
<affect>{space}+        ;

<affect>"FALSE"|"OFF"   return make_boolean (FALSE);
<affect>"TRUE"|"ON"     return make_boolean (TRUE);

<affect>{qstring}       return make_string ();
<affect>{identifier}    return make_identifier ();
<affect>{integer}       return make_integer ();
<affect>{real}          return make_real ();
<affect>"\n"            BEGIN (INITIAL);
<affect>.               ;

<<EOF>>                 yyterminate ();
%%

extern void *
yy_start_lex (const FILE *file, int size, const cfl_file_type_t file_type)
{
  YY_BUFFER_STATE nbuf = yy_create_buffer ((FILE *) file, size);
  yy_switch_to_buffer (nbuf);
  return (void *) nbuf;
}

extern void
yy_stop_lex (void *nbuf)
{
  yy_delete_buffer ((YY_BUFFER_STATE) nbuf);
}

static int make_string (void)
{
  yy_lval.string = (char *) malloc (yy_leng - 1);

  if (yy_lval.string != NULL)
  {
    strncpy (yy_lval.string, yy_text + 1, yy_leng - 2);
    yy_lval.string[yy_leng - 2] = '\0';
  }

#ifdef DEBUG
  printf ("STRING: <%s>\n", yy_lval.string);
#endif

  return STRING;
}

static int make_identifier (void)
{
  yy_lval.identifier = (char *) malloc (yy_leng + 1);

  if (yy_lval.identifier != NULL)
  {
    strncpy (yy_lval.identifier, yy_text, yy_leng);
    yy_lval.identifier[yy_leng] = '\0';
  }

#ifdef DEBUG
  printf ("IDENTIFIER: <%s>\n", yy_lval.identifier);
#endif

  return IDENTIFIER;
}

static int make_integer (void)
{
  long int e = strtol (yy_text, NULL, 0);
  yy_lval.integer = (long int *) malloc (sizeof (long int));

  if (yy_lval.integer != NULL )
  {
    memcpy (yy_lval.integer, &e, sizeof (long int));
  }

#ifdef DEBUG
  printf ("INTEGER: <%d>\n", e);
#endif

  return INTEGER;
}

static int make_real (void)
{
  double f = strtod (yy_text, NULL);
  yy_lval.real = (double *) malloc (sizeof (double));

  if (yy_lval.real != NULL)
  {
    memcpy (yy_lval.real, &f, sizeof (double));
  }

#ifdef DEBUG
  printf ("REAL: <%f>\n", f);
#endif

  return REAL;
}

static int make_boolean (bool val)
{
  yy_lval.boolean = (bool *) malloc (sizeof (bool));

  if (yy_lval.boolean != NULL)
  {
    memcpy (yy_lval.boolean, &val, sizeof (bool));
  }

#ifdef DEBUG
  printf ("BOOLEAN: <%s>\n", val == TRUE ? "TRUE" : "FALSE");
#endif

  return BOOLEAN;
}

