/*
 * This file is part of the Configuration File Library (CFL).
 * Copyright (C) 2002-2003 Nicolas Darnis <ndarnis@free.fr>
 *
 * The CFL library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The CFL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the CFL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: section.c
 * $Revision: 1.8 $
 * $Date: 2003/05/28 15:53:51 $
 * $Author: ndarnis $
 * $Id: section.c,v 1.8 2003/05/28 15:53:51 ndarnis Exp $
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <gdsl.h>

#include "cfl.h"
#include "section.h"

struct _section
{
  char       *section_name;
  gdsl_list_t section_values;
};

struct _value
{
  char *           value_name;
  cfl_value_type_t value_type;
  union
  {
    bool *    b;
    char *    s;
    long int *i;
    double *  d;
  } value;
};

static gdsl_element_t
value_alloc (void *arg);

static void
value_free (gdsl_element_t e);

static int
value_compare (gdsl_element_t e, void *val);

static void
value_write (gdsl_element_t e, FILE *file, void *user_data);

static void
value_write_xml (gdsl_element_t e, FILE *, void *user_data);

/******************************************************************************/
/* Public functions                                                           */
/******************************************************************************/

extern gdsl_element_t
section_alloc (void *arg)
{
  char *        section_name = (char *) arg;
  cfl_section_t s = NULL;  

  s = (cfl_section_t) malloc (sizeof (struct _section));

  if (s == NULL)
    {
      return NULL;
    }

  s->section_name = NULL;

  if (section_name != NULL)
    {
      s->section_name = (char *) malloc ((1 + strlen (section_name)) * sizeof (char));
      
      if (s->section_name == NULL)
	{
	  free (s);
	  return NULL;
	}

      strcpy (s->section_name, section_name);
    }

  s->section_values = gdsl_list_create ("VALUES", value_alloc, value_free);

  if (s->section_values == NULL)
  {
    if (s->section_name != NULL)
      {
	free (s->section_name);
      }

    free (s);
    return NULL;
  }

  return (gdsl_element_t) s;
}

extern void
section_free (gdsl_element_t e)
{
  cfl_section_t s = (cfl_section_t) e;

  if (s->section_name != NULL)
    {
      free (s->section_name);
    }

  gdsl_list_destroy (s->section_values);
  free (s);
}

extern int
section_compare (gdsl_element_t e, void *val)
{
  char *s1;
  char *s2;

  s1 = ((cfl_section_t) e)->section_name;
  s2 = (char *) val;

  if (s1 == NULL)
    {
      return (s2 == NULL) ? 0 : 1;
    }

  if (s2 == NULL)
    {
      return (s1 == NULL) ? 0 : 1;
    }

  return strcmp (s1, s2);
}

extern void
section_write (gdsl_element_t e, FILE *file, void *user_data)
{
  if (((cfl_section_t) e)->section_name == NULL)
    {
      gdsl_list_write (((cfl_section_t) e)->section_values, value_write, file, user_data);
      fprintf (file, "\n");
    }
  else
    {
      fprintf (file, "[%s]\n", ((cfl_section_t) e)->section_name);
      gdsl_list_write (((cfl_section_t) e)->section_values, value_write, file, user_data);
      fprintf (file, "\n");
    }
}

extern void
section_write_xml (gdsl_element_t e, FILE *file, void *user_data)
{
  cfl_section_t s = (cfl_section_t) e;

  if (s->section_name == NULL)
    {
      fprintf (file, "<CFL_BUILT_IN_SECTION VALUES=\"%d\">\n", 
	       gdsl_list_get_card (s->section_values));
    }
  else
    {
      fprintf (file, "<CFL_SECTION NAME=\"%s\" VALUES=\"%d\">\n", 
	       s->section_name, gdsl_list_get_card (s->section_values));
    }

  gdsl_list_write (s->section_values, value_write_xml, file, user_data);

  if (s->section_name == NULL)
    {
      fprintf (file, "</CFL_BUILT_IN_SECTION>\n");
    }
  else
    {
      fprintf (file, "</CFL_SECTION>\n");
    }
}

extern int
section_add_value (cfl_section_t s, const char *value_name, 
		   cfl_value_type_t type, void *val)
{
  struct _value v;
  cfl_value_t   tmp;

  tmp = (cfl_value_t) gdsl_list_search_by_value (s->section_values, 
						 value_compare, 
						 (void *) value_name);

  /* Value 'value_name' already exists in section 's' */
  if (tmp != NULL)
    {
      return CFL_ERR_VALUE_EXISTS;
    }

  v.value_name = (char *) value_name;
  v.value_type = type;
  
  switch (type)
    {
    case CFL_TYPE_STRING:
      v.value.s = val;
      break;
      
    case CFL_TYPE_INTEGER:
      v.value.i = val;
      break;
      
    case CFL_TYPE_FLOAT:
      v.value.d = val;
      break;

    case CFL_TYPE_BOOL:
      v.value.b = val;
      break;
    }

  if (gdsl_list_insert_tail (s->section_values, &v) == NULL)
    {
      return CFL_ERR_MALLOC;
    }

  return 0;
}

extern int
section_remove_value (cfl_section_t s, const char *value_name)
{
  gdsl_element_t e;

  e = gdsl_list_remove_by_value (s->section_values, value_compare, 
				 (void *) value_name);

  if (e == NULL)
    {
      return CFL_ERR_VALUE_NOT_EXISTS;
    }

  value_free (e);

  return 0;
}

extern int
section_set_value (cfl_section_t s, const char *value_name, cfl_value_type_t type, void *val)
{
  struct _value v;
  cfl_value_t   tmp;

  tmp = (cfl_value_t) gdsl_list_search_by_value (s->section_values, 
						 value_compare, 
						 (void *) value_name);

  /* Value not existing */
  if (tmp == NULL)
    {
      return CFL_ERR_VALUE_NOT_EXISTS;
    }

  switch (tmp->value_type)
    {
    case CFL_TYPE_STRING:
      if (tmp->value.s != NULL)
	{
	  free (tmp->value.s);
	  tmp->value.s = NULL;
	}
      break;
      
    case CFL_TYPE_INTEGER:
      if (tmp->value.i != NULL)
	{
	  free (tmp->value.i);
	  tmp->value.i = NULL;
	}
      break;
      
    case CFL_TYPE_FLOAT:
      if (tmp->value.d != NULL)
	{
	  free (tmp->value.d);
	  tmp->value.d = NULL;
	}
      break;
      
    case CFL_TYPE_BOOL:
      if (tmp->value.b != NULL)
	{
	  free (tmp->value.b);
	  tmp->value.b = NULL;
	}
      break;
    }

  tmp->value_type = type;
  
  switch (type)
    {
    case CFL_TYPE_STRING:
      tmp->value.s = (char *) malloc ((1 + strlen ((char *) val)) * sizeof (char));
      if (tmp->value.s == NULL)
	{
	  return CFL_ERR_MALLOC;
	}
      strcpy (tmp->value.s, (char *) val);
      break;
      
    case CFL_TYPE_INTEGER:
      tmp->value.i = (long int *) malloc (sizeof (long int));
      if (tmp->value.i == NULL)
	{
	  return CFL_ERR_MALLOC;
	}
      memcpy (tmp->value.i, (long int *) val, sizeof (long int));
      break;
      
    case CFL_TYPE_FLOAT:
      tmp->value.d = (double *) malloc (sizeof (double));
      if (tmp->value.d == NULL)
	{
	  return CFL_ERR_MALLOC;
	}
      memcpy (tmp->value.d, (double *) val, sizeof (double));
      break;
      
    case CFL_TYPE_BOOL:
      tmp->value.b = (bool *) malloc (sizeof (bool));
      if (tmp->value.b == NULL)
	{
	  return CFL_ERR_MALLOC;
	}
      memcpy (tmp->value.b, (bool *) val, sizeof (bool));
      break;
    }
  
  return 0;
}

extern void *
section_get_value (cfl_section_t s, const char *value_name, cfl_value_type_t *type)
{
  cfl_value_t v;

  v = (cfl_value_t) gdsl_list_search_by_value (s->section_values, value_compare, 
					       (void *) value_name);

  if (v == NULL)
    {
      return NULL;
    }

  if (type)
    {
      *type = v->value_type;
    }

  switch (v->value_type)
    {
    case CFL_TYPE_STRING:
      return v->value.s;

    case CFL_TYPE_INTEGER:
      return v->value.i;

    case CFL_TYPE_FLOAT:
      return v->value.d;

    case CFL_TYPE_BOOL:
      return v->value.b;
    }

  return NULL;
}

extern void *
section_get_value_by_pos (cfl_section_t s, ulong value_pos, char **value_name, 
			  cfl_value_type_t *type)
{
  cfl_value_t v;

  v = (cfl_value_t) gdsl_list_search_by_position (s->section_values, value_pos);

  if (v == NULL)
    {
      return NULL;
    }

  if (type != NULL)
    {
      *type = v->value_type;
    }

  if (value_name)
    {
      *value_name = v->value_name;
    }

  switch (v->value_type)
    {
    case CFL_TYPE_STRING:
      return v->value.s;

    case CFL_TYPE_INTEGER:
      return v->value.i;

    case CFL_TYPE_FLOAT:
      return v->value.d;

    case CFL_TYPE_BOOL:
      return v->value.b;
    }

  return NULL;
}

extern ulong
section_get_values_number (cfl_section_t s)
{
  return gdsl_list_get_card (s->section_values);
}

extern char *
section_get_name (cfl_section_t s)
{
  return s->section_name;
}

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static gdsl_element_t
value_alloc (void *arg)
{
  cfl_value_t v;
  cfl_value_t tmp = (cfl_value_t) arg;

  v = (cfl_value_t) malloc (sizeof (struct _value));

  if (v == NULL)
    {
      return NULL;
    }

  v->value_name = (char *) malloc ((1 + strlen (tmp->value_name)) * sizeof (char));

  if (v->value_name == NULL)
    {
      free (v);
      return NULL;
    }

  strcpy (v->value_name, tmp->value_name);

  v->value_type = tmp->value_type;

  switch (v->value_type)
    {
    case CFL_TYPE_STRING:
      v->value.s = (char *) malloc ((1 + strlen (tmp->value.s)) * sizeof (char));

      if (v->value.s == NULL)
	{
	  free (v->value_name);
	  free (v);
	  return NULL;
	}

      strcpy (v->value.s, tmp->value.s);
      break;

    case CFL_TYPE_INTEGER:
      v->value.i = (long int *) malloc (sizeof (long int));

      if (v->value.i == NULL)
	{
	  free (v->value_name);
	  free (v);
	  return NULL;
	}

      memcpy (v->value.i, (long int *) tmp->value.i, sizeof (long int));
      break;

    case CFL_TYPE_FLOAT:
      v->value.d = (double *) malloc (sizeof (double));

      if (v->value.d == NULL)
	{
	  free (v->value_name);
	  free (v);
	  return NULL;
	}

      memcpy (v->value.d, (double *) tmp->value.d, sizeof (double));
      break;

    case CFL_TYPE_BOOL:
      v->value.b = (bool *) malloc (sizeof (bool));

      if (v->value.b == NULL)
	{
	  free (v->value_name);
	  free (v);
	  return NULL;
	}

      memcpy (v->value.b, (bool *) tmp->value.b, sizeof (bool));
      break;
    }

  return v;
}

static void
value_free (gdsl_element_t e)
{
  cfl_value_t v = (cfl_value_t) e;

  switch (v->value_type)
    {
    case CFL_TYPE_STRING:
      if (v->value.s != NULL)
	{
	  free (v->value.s);
	}
      break;
      
    case CFL_TYPE_INTEGER:
      if (v->value.i)
	{
	  free (v->value.i);
	}
      break;

    case CFL_TYPE_FLOAT:
      if (v->value.d)
	{
	  free (v->value.d);
	}
      break;

    case CFL_TYPE_BOOL:
      if (v->value.b)
	{
	  free (v->value.b);
	}
      break;
    }

  free (v->value_name);
  free (v);
}

static int
value_compare (gdsl_element_t e, void *val)
{
  return strcmp (((cfl_value_t) e)->value_name, (char *) val);
}

static void
value_write (gdsl_element_t e, FILE *file, void *user_data)
{
  cfl_value_t v = (cfl_value_t) e;

  switch (v->value_type)
    {
    case CFL_TYPE_STRING:
      fprintf (file, "%s = \"%s\"\n", v->value_name, v->value.s);
      break;
      
    case CFL_TYPE_INTEGER:
      fprintf (file, "%s = %ld\n", v->value_name, *(v->value.i));
      break;
      
    case CFL_TYPE_FLOAT:
      fprintf (file, "%s = %g\n", v->value_name, *(v->value.d));
      break;

    case CFL_TYPE_BOOL:
      fprintf (file, "%s = %s\n", v->value_name, *(v->value.b) ? "TRUE" : "FALSE");
      break;
    }
}

static void
value_write_xml (gdsl_element_t e, FILE *file, void *user_data)
{
  cfl_value_t v = (cfl_value_t) e;

  fprintf (file, "<CFL_VALUE NAME=\"%s\" ", v->value_name);

  switch (v->value_type)
    {
    case CFL_TYPE_STRING:
      fprintf (file, "TYPE=\"STRING\" VALUE=\"%s\"/>\n", v->value.s);
      break;
      
    case CFL_TYPE_INTEGER:
      fprintf (file, "TYPE=\"INTEGER\" VALUE=\"%ld\"/>\n", *(v->value.i));
      break;
      
    case CFL_TYPE_FLOAT:
      fprintf (file, "TYPE=\"FLOAT\" VALUE=\"%g\"/>\n", *(v->value.d));
      break;

    case CFL_TYPE_BOOL:
      fprintf (file, "TYPE=\"BOOL\" VALUE=\"%s\"/>\n", *(v->value.b) ? "TRUE" : "FALSE");
      break;
    }
}

